/**
 *  \file telnet.c
 *
 *  weAutSys Telnet server implementation
 *
 *  This is system software and must not be modified for user or
 *  application programs.
 *
 * This file is part of weAutSys &nbsp;
 *           <<a href="http://weinert-automation.de">weinert-automation.de</a>>
 *
 * &nbsp; \ref copyLicNote " Copyright &copy; " 2012 Albrecht Weinert, Bochum
 * \author  Albrecht Weinert  &nbsp; <<a href="http://a-weinert.de">a-weinert.de</a>>
 * $Revision: 2 $
 * $Date: 2017-01-25 16:49:03 +0000 (Wed, 25 Jan 2017) $
 */
//  SVN last change by $Author: albrecht $

#if defined(arduMega2560)
   // no network dummy
#elif defined(arduinoUno)
   // no network dummy
#else // defined(weAut_00) || defined(weAut_01)

#include "we-aut_sys/common.h"
#include "we-aut_sys/telnet.h"
#include "we-aut_sys/streams.h"
#include "we-aut_sys/log_streams.h"
#include "we-aut_sys/system.h"
#include "we-aut_sys/utils.h"
#include "we-aut_sys/cli.h"

#include "uip/uip.h"

#include <string.h>

#if DEBUG_TEL
// #include "we-aut_sys/uart0.h"

//                            0123456789.
INFLASH(char const telnMark[])   = " # Telnet ";
INFLASH(char const telnDO[])     = " # TelnDO ";
INFLASH(char const telnWILL[])   = " # TelnWI ";
INFLASH(char const telnDONT[])   = " # TelnDN ";
INFLASH(char const telnWONT[])   = " # TelnDN ";
INFLASH(char const telnSB[])     = " # TelnSB ";
#endif

/*  Initialise the Telnet server
 *
 *  This function initialises the Telnet server. It is (indirectly) called
 *  whenever the Ethernet link goes up and hence usually after reset.
 *  This implementation does nothing.
 */
void telnetInit(void){ }


/*  Handle Telnet  events
 *
 *  This function handles Telnet events. It is a as a protothread (part of)
 *  the Telnet server state machine.
 *
 *  Hint: This implements at at present the basic communication and a CLI
 *  (command line interpreter).
 *
 *  Hint 2: for debugging (deprecated) UART logging is still used instead
 *  of bufLog...(). As the buffered logger may connect to a Telnet session
 *  Telnet would be used to debug itself (should that become necessary in
 *  future).s
 *
 *  The current appstate.flag usage is: <br />
 *  Bit 0 (1): command interpreter active <br />
 *  Bit 1 (2): prompt output (after last output) pending <br />
 *  Bit 3 (4): consume and send the pending \ref bufLogStreams output
 *           (before prompt). <br />
 */
ptfnct_t telnetAppcall(void){

   struct thr_data_t *s = (struct thr_data_t *)&(uip_conn->appstate);

   if (uip_connected()) { // newly connected  ----------------------------
#if DEBUG_TEL
      //           0123456789.123456789v123456 789t
      char tS[] = " ## Telnet conn .... .... \n";
      fourHexs(&tS[16], (uint16_t)bufLogCons);
      fourHexsBE(&tS[21], uip_conn->rport);
      bufLogTxt(tS, 27);
#endif
      initAsCLIthread(s, &lanStreams);
      s->cliThrData.commNumb = // command version greeting as system command
            comVersionNum + INDEX_OFFSET_LIST2; // simulate command entry
      s->flag = 3; // output prompt on next no command poll

      if (!bufLogCons) bufLogCons = s; // auto logging on if none


      //  wrong now   if (telnetLogBuf(0) == UIP_CONNS) s->flag = 0x83;
      return PT_EXITED;
   } // ---------------     newly connected   -------------------- end ----

   if (uip_closed()) { // closed by other host (client)   -----------------
      // uip_close();  // not necessary
#if DEBUG_TEL
         bufLogHex_P(telnMark, 0xFC); // foreign client has closed
#endif
         PT_LEAVE(&s->pt);
   }  // --------------   closed by other host (client)  ---------- end ----

   if (uip_aborted()) { // aborted by other host (client) ------------------
      uip_close();  // not necessary
#if DEBUG_TEL
         bufLogHex_P(telnMark, 0xFA); // foreign client has aborted
#endif
         PT_LEAVE(&s->pt);
   }  // -------------     aborted by other host (client) --------- end ----

   if (uip_newdata() ) {  //= = = = = = = = = = =  =  new  data  begin  ====
      const uint16_t gotLen = uip_datalen();
      if (!gotLen) {
         return PT_EXITED;
      }
      lanStreamsInit();

      // todo better criteria  we search for any FF anywhere
      //      a) filter FF FF
      //      b) restrict command to start of message (normally is)
      // if (memchr((char *)uip_appdata, 255, gotLen)) { // has IAC
      // we do b)

      if (*((char *)uip_appdata) == IAC) { // has Telnet IAC (command) =====
#if DEBUG_TEL
         bufLogHex_P(telnMark, 0xAC); // arrived command
#endif
         // Current tactics on Telnet IAC commands s:
         //                    refuse all and hope for (kludge) line mode
         char * readp = (char *)uip_appdata;
         uint8_t iacState = 0;
         uint8_t index = 0;
         while (index++ < gotLen) {
            uint8_t c = *(readp++);
            switch (iacState ) {
               case WILL : // want or confirm indicated option
                  lanPutC(IAC);
#if DEBUG_TEL
                  bufLogHex_P(telnWILL, c);
#endif
                  //   lanPutC(c == LINEMODE ? DO : DONT);
                  //   lanPutC(c == SUPPRESS_GA ? DO : DONT);
                  lanPutC( DONT);
                  lanPutC(c);
                  iacState  = 0;
                  break;

               case WONT : // refuse or continue indicated option
#if DEBUG_TEL
                  bufLogHex_P(telnWONT, c);
#endif

                  lanPutC(IAC);
                  lanPutC(DONT); // c == LINEMODE  is a problem
                  lanPutC(c);
                  iacState  = 0;
                  break;

               case  DO :  // request indicated option
#if DEBUG_TEL
                  bufLogHex_P(telnDO,  c);
#endif

                  lanPutC(IAC);
                  //    lanPutC(c == LINEMODE  ? WILL : WONT);
                  //     lanPutC(c == SUPPRESS_GA ? WILL : WONT);
                  lanPutC(WONT);
                  lanPutC(c);
                  iacState  = 0;
                  break;

               case  DONT : // demand (other) no longer to perform indicated option
#if DEBUG_TEL
                  bufLogHex_P(telnDONT, c);
#endif
                  lanPutC(IAC);
                  lanPutC(WONT); // c == LINEMODE  is a problem
                  lanPutC(c);
                  iacState  = 0;
                  break;

               case  TEL_SB :  // in sub-negotiation
#if DEBUG_TEL
                  bufLogHex_P(telnSB, c);
#endif

                  if (c == IAC) iacState  = IAC;
                  // ignore all other values (we do no sub-negotiation yet
                  break;

               case IAC :
                  iacState  = (c == IAC || c == TEL_SE) ? 0 : c;
                  break;

               default: // normal
                  iacState = c == IAC ? IAC : 0;
                  // ignore all other values
                  // (we do no accept text mixed in a line with commands)
                  break;
            } // case state

         } // while through input

         lanStreamSend(); // also closes
         return PT_EXITED; // had Telnet command (IAC) -> out
      } // (first) char was IAC
//--  to here with normal (text) message / end had Telnet IAC (command) ====

#if DEBUG_TEL
      bufLogHex_P(telnMark, 0xAD); // arrived data
#endif
      if (setCliLine(&s->cliThrData, (char *)uip_appdata,
            uip_datalen()) == 255) { // parse line and prep. CLI
         lanStreamsEcho();
         stdPutS(" ?? try help \n \nweAutSys Telnet > ");
         s->cliThrData.commNumb = 0;
         s->flag &= ~3; // clear command and prompt bits
      } else {
         if (s->cliThrData.commNumb != (comTelnetNum
               + INDEX_OFFSET_LIST2)) { // exclude Telnet Command here
            sysCliThreadF(&s->cliThrData); // let one (first) round CLI work
         } // exclude Telnet command to be handled in next poll phase
         // output prompt on next no command poll
         s->flag  |=  ( s->cliThrData.commNumb ? 3 : 2);
#if DEBUG_TEL
         //        0123456789.123456789v12 3
         char tS[] = " ## Telnet CLIa .. .. \n";
         twoHexs(&tS[16], s->cliThrData.commNumb);
         //  twoHexs(&tS[19], s->cliThrData.outFlashTextThread.index);
         twoHexs(&tS[19], s->flag);
         bufLogTxt(tS, 23);
#endif
      }
      lanStreamSend(); // also closes

      PT_EXIT(&s->pt);
   } //= = = = = = = = = = = = = = = = = = = = =  =  new  data  (end)  ====


   if (uip_poll() ) { //= = = = = = = = = = = = = = =   poll     begin ====
      if (!s->flag) {
         if (s == bufLogCons && bufLogInBufferd(NULL)) {
            s->flag = 4;  // log output pending
         } else {
            PT_LEAVE(&s->pt);  // end normally
         }
      }

      PT_BEGIN(&s->pt);  // standard schedule (re-) entry point of thread
      lanStreamsInit();

#if DEBUG_TEL
      //        0123456789.123456789v12345 6
      char tS[] = " ## TelnetCpoll .. .. ...\n";
      twoHexs(&tS[16], s->cliThrData.commNumb);
      // twoHexs(&tS[19], s->cliThrData.outFlashTextThread.index);
      twoHexs(&tS[19], s->flag);
      threeDigs(&tS[22], outSpace(stdout));
      bufLogTxt(tS, 26);
#endif
      if (s->cliThrData.commNumb == (comTelnetNum + INDEX_OFFSET_LIST2)) {
         s->cliThrData.commNumb = 0; // consider command done
         switch (s->cliThrData.optionNumb) {

            case optAbortNum:  // Telnet abort
               uip_abort();
               goto telnetEnd;

            case optOffNum: // Telnet off
            case optStopNum:
            case optResetNum:
               uip_close();
               telnetEnd:
               if (s == bufLogCons) bufLogCons = NULL;
               PT_LEAVE(&s->pt);  // end normally
               break;

            case optLogHereNum:  // telnet -logHere
               bufLogCons = s;
               s->cliThrData.commNumb = 0;
               if (bufLogInBufferd(NULL)) {
                  s->flag = 6; // log and prompt
               } else {
                  s->flag = 2; // prompt
               }
               break;

            case optNoLogsNum: // telnet -noLogs
               if (s == bufLogCons) bufLogCons = NULL;
               break;

            case optNotGivenNum:
               break;
            default:
               unimplOptionReport(&(s->cliThrData)); // helper
               break; // no option or not used here
         } // option switch
      } // telnet command processing

      // normal command processing
      while (s->cliThrData.commNumb) { // command to be done
         const uint8_t cliRet = sysCliThreadF(&s->cliThrData);
         if(cliRet >= PT_EXITED) break;
         lanStreamSend(); // also closes
         return PT_YIELDED;
      } // command to be done

      if (!(s->cliThrData.commNumb)) { // command is done
         s->flag &= ~1; // no more Command output
         if (s == bufLogCons && bufLogInBufferd(NULL)) { // telnet : log output
            const uint16_t tfc = bufLogGetChars(
                  ((char *)uip_appdata + uip_conn->writeIndex),
                                                 lanStreamsOutSpace(NULL));
            const uint16_t remain =  bufLogInBufferd(NULL);

#if DEBUG_TEL
         //        0123456789.123456789v1234 56
      char tS[] = " ## TelnetBufC ... / ...\n";
      threeDigs(&tS[15], tfc);
      threeDigs(&tS[21], remain);
      bufLogTxt(tS, 25);
#endif

            if (!remain) s->flag &= ~4; // no more log output
            if (tfc) { // some output
               uip_conn->writeIndex += tfc;
               lanStreamSend(); // also closes
               PT_LEAVE(&s->pt);  // end normally
            } // had log output
            /// PT_YIELD_OUT_SPACE(&s->pt, 33, &lanStreams);
            ///stdPutS(" \n ## test simulation of \n"
            ///   " ## log buffer output ");
         } else { // this telnet is log output
            s->flag &= ~4; // no more log output
         }


         if (s->flag == 2  // make prompt
               && lanStreamsOutSpace(NULL) > 23) { // space

            stdPutS(" \nweAutSys Telnet > ");
            s->flag &= 0x80;
         } // make prompt & space

      } // command is done

      lanStreamSend(); // also closes
      PT_END(&s->pt)  // standard schedule (re-) entry point of thread
   } //= = = = = = = = = = = = = = = = = = = = =  =  poll     (end)  ====

   // neither connected nor newData, poll or closed
   // what else ? uip_aborted()
   return PT_EXITED;
} // telnetAppcall()

#endif // no ethernet
