/**
 *  \file timing.c
 *
 *  weAutSys  implementation of timing services
 *
 *  This is system software and must not be modified for user or
 *  application programs.
 *
 *  The file timing.c is part of weAutSys &nbsp;
 *           <<a href="http://weinert-automation.de">weinert-automation.de</a>>
 *
 * &nbsp; \ref copyLicNote " Copyright &copy; " 2011 Albrecht Weinert, Bochum
 * \author  Albrecht Weinert  &nbsp; <<a href="http://a-weinert.de">a-weinert.de</a>>
 * $Revision: 2 $
 * $Date: 2017-01-25 16:49:03 +0000 (Wed, 25 Jan 2017) $
 */
//  SVN last change by $Author: albrecht $

#include "we-aut_sys/common.h"
#include "we-aut_sys/timing.h"
#include "we-aut_sys/utils.h"
#include "we-aut_sys/syst_threads.h"

#include "we-aut_sys/log_streams.h"


// for unix ntp time calculations:    http://www.aelius.com/njh/unixtime/
// for DST switches: http://www.timeanddate.com/worldclock/timezone.html?n=37

/*  Milliseconds since reset / power up.
 *
 *  This are the counted milliseconds since reset, start or power up.
 *  The value will wrap around in about 49 days.
 * */
uint32_t msClock(void) { return msTimers.end.v32; }

/*  The system's run time in seconds
 *
 *  These are the seconds since reset, start or power up. The value will
 *  wrap around in about 136 years.
 */
uint32_t secClock(void) { return secDurTimers.end.v32; }

/*  The local time as FAT time
 *
 *  The local time contains zone offsets and \ref isDST "summertime (DST)"
 *  shifts. This function returns this time in a 32 bit (uint32_t) packed
 *  structure used for the traditional FAT file systems. Note that this
 *  (old DOS) time format has a 2s resolution only. The format is:
 *
 *  bit31..25: year from 1980 (0..127; 2012 is 32) <br />
 *  bit24..21: month (1..12) <br />
 *  bit20..16: day in month (1..31) <br />
 *  bit15..11: hour (0..23) <br />
 *  bit10...5: minute (0..59) <br />
 *  bit 4...0: second / 2 (0..29, 30 if leap second)
 *
 *  \sa secTime308Loc
 */
uint32_t getFATtime(void){
   ucnt32_t dur;
   dur.v32 =  secTime308Loc();
   datdur_t  sysTim;
   sec2datdur(&sysTim, dur.v32);
   date_t sysDat;
   setDatByDays(&sysDat, sysTim.d);

   dur.v8[3] = ((sysDat.y + 20) << 1);
   if (sysDat.m > 7) dur.v8[3] |= 1;
   dur.v8[3] = (sysDat.m << 5) | sysDat.d;

   dur.v8[2] = (sysTim.h << 3) | (sysTim.m >> 3);
   dur.v8[2] = (sysTim.m << 5) | (sysTim.s >> 1);
   return dur.v32;
} // getFATtime()


/* The combined offset (local - UTC)
 *
 *  This is the difference from  \ref secTime308Loc "local time" to
 *  \ref secTime308UTC "UTC".<br />
 *  In other words it is the sum \ref zoneOffsetSec + \ref setDST "DST offset"
 */
uint32_t combinedOffset;

/* Is current time DST */
uint8_t isDST;

/* Set if current time is  DST  */
uint8_t setDST(uint8_t dlt) {

   uint32_t newCo;
   if (dlt) { // set DLT on
      if (isDST) return 0;
      isDST = 1;
      newCo = zoneOffsetSec + 3600;
   } else {
      if (!isDST) return 0;
      isDST = 0;
      newCo = zoneOffsetSec;
   }
   uint32_t dif =  newCo - combinedOffset;
   if (!dif) return 0;
   secDatTimers.end.v32 += dif;
   sys1sPeriodFlag |= ABSsecADJ | ABSsecSET;
   combinedOffset = newCo;
   return 1;
} // setDST(uint8_t

/* Seconds since January 1st 1970 UTC (Unix time) */
uint32_t secTimeUnixUTC(uint32_t secTime308UTC){
   return secTime308UTC + OFFSET2UNIXTIME;
} // secTimeUnixUTC(void)

/* Seconds since January 1st 1990 UTC (NTP time) */
uint32_t secTimeNtpUTC(uint32_t secTime308UTC){
   return secTime308UTC + OFFSET2NTPTIME;
} // secTimeNtpUTC(void)


/* Seconds since March 1st 2008 UTC  */
uint32_t secTime308UTC(void){
   return secDatTimers.end.v32 - combinedOffset;
} // secTime308UTC(void)


/* Adjust / set zone offset
 *
 *  This function sets \ref zoneOffsetSec and handles all side effects in the
 *  case of a change.
 *  \return true if changed
 */
uint8_t setZoneOffsetSec(uint32_t newOffset){
   uint32_t dif = newOffset - zoneOffsetSec;
   if (!dif) return 0;
   zoneOffsetSec = newOffset;
   combinedOffset += dif;
   secDatTimers.end.v32 += dif;
   return 1;
} // setZoneOffsetSec(uint32_t newOffset)

uint8_t adjustUTCcount;

/* The system milliseconds time and timers
 *
 *  This is (a pseudo timer) for system internal timing use only.
 *  Application software must not modify any field.
 *
 *  \c  msTimers.next is the root of the system's internal ordered linked
 *  list of active / running ms based timers: it is either NULL or points
 *  to the first ("quickest")  running timer.
 *
 *  \c  msTimers.end is the (one) system ms 32 bit counter.
 *
 *  \c  msTimers.period is the time zone offset in seconds (3600 for CET,
 *  7200 for CEST).
 */
struct timer_t msTimers;

/* The system seconds resolution date / time oriented timers
 *
 *  This is (a pseudo timer) for system internal timing  and list handling
 *  use only!
 *
 *  \c .next is the root of the system's internal ordered linked
 *  list of active / running seconds based timers.
 *
 *  \c  secDatTimers.end  is the local time in seconds since
 *  1st 2008 UTC.
 *
 *  \c  secDatTimers.period is the (\c msTimers.end) value of the absolute
 *  time's last adjustment or setting.
 */
struct timer_t secDatTimers;

/* The system seconds resolution duration oriented timers
 *
 *  This is (a pseudo timer) for system internal timing  and list handling
 *  use only!
 *
 *  \c .next is the root of the system's internal ordered linked
 *  list of active / running seconds based timers.
 *
 *  \c  secDurTimers.end is the (one) system seconds clock counting the
 *  seconds since last reset / re-start.
 *
 *  \c  secDurTimers.period is the system start time in
 *  seconds since March 1st 2008 UTC.
 */
struct timer_t secDurTimers;

/* The list of timer lists
 *
 *  Index is the type: ms_TIMER_TYPE,  sec_dur_TIMER_TYPE
 *  see sec_dat_TIMER_TYPE TIMER_TYPE_MASK
 */
struct timer_t * timerLists[] = {&msTimers, &secDatTimers, &secDurTimers, NULL};

/* Remove a timer from a list (internal helper)
 *
 * \param timer the timer to be removed
 * \param list the list head, but may be any timer (in a list) the search
 *        for timer to be remover will start after
 */
struct timer_t * removeTimerFromList(const struct timer_t * timer,
                                                     struct timer_t * list){
   struct timer_t * nxInL = list->next;  // next in list
   while (nxInL != NULL) {
      if (nxInL == timer) {
         list->next = nxInL->next;
         nxInL->next = NULL;
         return list;
      }
      list = nxInL;
      nxInL = nxInL->next;
   } // nxInL != null
   return NULL;
} //  removeTimerFromList(struct timer_t * timer, timer_t * list)


/*  Start a timer with interval / type  from now
 *
 *  \param timer points to the timer to be re-started
 *  \param period is the next interval duration in s or ms depending on \c type
 *  \param type may be \ref ms_TIMER_TYPE,  \ref  sec_dur_TIMER_TYPE or
 *              \ref  sec_dat_TIMER_TYPE; the
 *              \ref  TIMER_STATE_MASK "upper 4 state" bits are ignored
 */
void startTimer(struct timer_t * timer, uint32_t period, uint8_t type){
   if (timer == NULL) return;
   const uint8_t oldType = timer->state & TIMER_TYPE_MASK;
   if (oldType == TIMER_TYPE_FREE)  return;
   const uint8_t oldState = timer->state & TIMER_STATE_MASK;
   if (oldState == TIMER_RUNNING) {
      removeTimerFromList(timer, timerLists[oldType]);
   } else {
      timer->next  = NULL;
   }
   timer->period.v32 = period;
   if ((type &= TIMER_TYPE_MASK) == TIMER_TYPE_DEFAULT) type = ms_TIMER_TYPE;

   if (period == 0) {
      timer->state = TIMER_UNUSED | type;
      return;
   }

   struct timer_t * before = timerLists[type];
   uint32_t endTime = timer->end.v32 =  before->end.v32 + period;
   timer->state = TIMER_RUNNING | type;

   while (1) {
      struct timer_t * after  = before->next;
      if (after == NULL || geU32ModAr(after->end.v32, endTime)) { // put b-a
         before->next = timer;
         timer->next = after;
         return;
      } // put before after
      before = after;
    } // while
   // to here never as after has to get NULL anyway
 } //  startTimer(struct timer_t * timer, uint32_t period, uint8_t type);


/*  Start a timer with its own interval from now
 *
 *  \param timer points to the timer to be re-started
 */
void restartTimer(struct timer_t * timer){
   if (timer == NULL) return;
   const uint8_t oldType = timer->state & TIMER_TYPE_MASK;
   if (oldType == TIMER_TYPE_FREE)  return;
   struct timer_t * before = NULL;
   const uint8_t oldState = timer->state & TIMER_STATE_MASK;
   if (oldState == TIMER_RUNNING) {
      before = removeTimerFromList(timer, timerLists[oldType]);
      // internal: timer can only advance its position in the (run) list
   } else {
      timer->next  = NULL;
   }
   if (timer->period.v32 == 0) {
      timer->state = TIMER_UNUSED | oldType;
      return;
   }

   if (!before) before = timerLists[oldType];
   uint32_t endTime =  before->end.v32 + timer->period.v32;
   timer->end.v32 = endTime;
   timer->state = TIMER_RUNNING | oldType;

   while (1) {
      struct timer_t * after  = before->next;
      if (after == NULL || geU32ModAr(after->end.v32, endTime)) { // put b-a
         before->next = timer;
         timer->next = after;
         return;
      } // put before after
      before = after;
    } // while
   // to here never as after has to get NULL anyway
 } // restartTimer(struct timer_t *);


/*  Start a timer (with absolute end time)
 *
 *  If timer is NULL or it is \ref TIMER_TYPE_FREE "free" nothing will
 *  happen.<br />
 *  If the \c endTime set is not in the future the state will be set to
 *  \ref TIMER_UNUSED "unused" and nothing else will happen.
 *
 *  Otherwise this function will set the \c timer
 *  \ref TIMER_RUNNING "running" to the \c endTime ms or s, no matter what
 *  its before state has been.
 *
 *  \param timer points to the timer to be started
 *  \param endTime is the next interval's end in s or ms depending on \c type
 *  \param type may be \ref ms_TIMER_TYPE,  \ref  sec_dur_TIMER_TYPE or
 *              \ref  sec_dat_TIMER_TYPE; the
 *              \ref  TIMER_STATE_MASK "upper 4 state" bits are ignored
 */
void startTimerAbs(struct timer_t * timer, uint32_t endTime, uint8_t type){
   if (timer == NULL) return;
   const uint8_t oldType = timer->state & TIMER_TYPE_MASK;
   if (oldType == TIMER_TYPE_FREE)  return;
   const uint8_t oldState = timer->state & TIMER_STATE_MASK;
   if (oldState == TIMER_RUNNING) {
      removeTimerFromList(timer, timerLists[oldType]);
   } else {
      timer->next  = NULL;
   }

   timer->end.v32 = endTime;
   if ((type &= TIMER_TYPE_MASK) == TIMER_TYPE_DEFAULT) type = ms_TIMER_TYPE;

   struct timer_t * before = timerLists[type];
   const uint32_t curTime = before->end.v32;
   if (geU32ModAr(curTime, endTime)) {
      timer->state = TIMER_UNUSED | type;
      return;
   }

   while (1) {
      struct timer_t * after  = before->next;
      if (after == NULL || geU32ModAr(after->end.v32, endTime)) { // put b-a
         before->next = timer;
         timer->next = after;
         timer->state = TIMER_RUNNING | type;
         return;
      } // put before after
      before = after;
    } // while
   // to here never as after has to get NULL anyway

} //  startTimerAbs(struct timer_t * timer, uint32_t endTime, uint8_t type


/* Start a timer's next interval
 *
 *  Start or restart a timer with its actually set internal \c period
 *  from its actual end time.<br />
 *  This actual end time may be in future for a running timer. In this case
 *  the next timer lapse will be postponed farther in the future.<br />
 *  The actual end time may be now or in the past for a lapsed timer.
 *  In this case it is a "frequency exact" restart, while
 *  \c restartTimer(struct timer_t) would be an "absolute time exact" restart.<br />
 *  Attention: If the the timers actual end time so far in the past that the
 *  prolonged endtime is in the past too, the timer will not be re-startet.
 *
 *  For timers that are neither running nor lapsed, this function will,
 *  start the timer from now.
 *
 * \param timer points to the timer to be re-started
 */
void startNextInterval(struct timer_t * timer){
   if (timer == NULL) return;
   const uint8_t oldType = timer->state & TIMER_TYPE_MASK;
   if (oldType == TIMER_TYPE_FREE)  return;
   uint32_t timerPeriod = timer->period.v32;
   if (timerPeriod == 0) return; // do nothing on period 0

   struct timer_t * before = NULL;
   uint8_t oldState = timer->state & TIMER_STATE_MASK;
   if (oldState == TIMER_RUNNING) {
      if (timer->next == NULL) { // last or only timer in the list
         timer->end.v32 += timerPeriod;   // so just prolong
         return;
      } // last or only timer in the list
      before = removeTimerFromList(timer, timerLists[oldType]);
      // internal: timer can only advance its position in the (run) list
      oldState = TIMER_LAPSED; // sets relative to false
   } else {
      timer->next  = NULL;
   }

   timer->state = TIMER_RUNNING | oldType;
   uint32_t endTime;

   if (oldState) { // absolute for all but running or lapsed
      timer->end.v32 = endTime = timerLists[oldType]->end.v32 + timerPeriod;
   } else { // relative
      timer->end.v32 = endTime = timer->end.v32 + timerPeriod;
     if (geU32ModAr( timerLists[oldType]->end.v32 , endTime)) {
         timer->state =  oldType | TIMER_LAPSED; // lapsed from start
         return;
      }
   }

   if (!before) before = timerLists[oldType];
   while (1) {
      struct timer_t * after  = before->next;
      if (after == NULL || geU32ModAr(after->end.v32, endTime)) { // put b-a
         before->next = timer;
         timer->next = after;
         return;
      } // put before after
      before = after;
    } // while
   // to here never as after has to get NULL anyway
} //  startNextInterval(struct timer_t * timer)

/* Halt a timer
 *
 * Just stop the timer. This will set the timer to state \c TIMER_UNUSED
 */
void haltTimer(struct timer_t * timer){
   if (timer == NULL) return;
   const uint8_t oldType = timer->state & TIMER_TYPE_MASK;
   if (oldType == TIMER_TYPE_FREE)  return;
   const uint8_t oldState = timer->state & TIMER_STATE_MASK;
   if (oldState == TIMER_RUNNING) {
      removeTimerFromList(timer, timerLists[oldType]);
   }
   timer->next  = NULL;
   timer->state = oldType | TIMER_UNUSED;
} // haltTimer(struct timer_t *)


/* Stop / pause a timer
 *
 *  A bit like \c haltTimer(struct timer_t), but the internal \c period will be
 *  reduced to the actual run duration not yet elapsed and the state will
 *  be TIMER_STOPPED not TIMER_UNUSED, if it was running before.
 *
 *  \c restartTimer(struct timer_t * timer) can be used afterwards to get it running
 *  on.<br />
 *  \c startTimer(struct timer_t * timer, uint32_t period) has to be used to
 *  restore the original state.<br />
 */
void pauseTimer(struct timer_t * timer){
   if (timer == NULL) return;
   const uint8_t oldType = timer->state & TIMER_TYPE_MASK;
   if (oldType == TIMER_TYPE_FREE)  return;
   const uint8_t oldState = timer->state & TIMER_STATE_MASK;
   if (oldState == TIMER_STOPPED) return;
   if (oldState == TIMER_RUNNING) {
      removeTimerFromList(timer, timerLists[oldType]);
      timer->period.v32 = timer->end.v32 - timerLists[oldType]->end.v32; // rest period
      timer->state = oldType |  TIMER_STOPPED;
   } else {
   timer->next  = NULL;
   timer->state = oldType | TIMER_UNUSED;
   }
} // pauseTimer(struct timer_t *)

/* Initialise a timer */
uint8_t initTimer(struct timer_t * timer, uint32_t period, uint8_t type){
   if (timer == NULL) return 0;
   const uint8_t oldType  = timer->state & TIMER_TYPE_MASK;
   const uint8_t oldState = timer->state & TIMER_STATE_MASK;
   if (oldState == TIMER_RUNNING && oldType != TIMER_TYPE_FREE) {
      removeTimerFromList(timer, timerLists[oldType]);
   } else {
      timer->next  = NULL;
   }
   timer->period.v32 = period;
   if ((type &= TIMER_TYPE_MASK) == TIMER_TYPE_DEFAULT) type = ms_TIMER_TYPE;
   timer->state = TIMER_UNUSED | type;
   return oldType == TIMER_TYPE_FREE ? TIMER_TYPE_FREE : 0;
} //  initTimer(struct timer_t *)



/* Free a timer  */
void freeTimer(struct timer_t * timer){
   if (timer == NULL) return;
   const uint8_t oldType = timer->state & TIMER_TYPE_MASK;
   if (oldType == TIMER_TYPE_FREE)  return;
   const uint8_t oldState = timer->state & TIMER_STATE_MASK;
   if (oldState == TIMER_RUNNING) {
      removeTimerFromList(timer, timerLists[oldType]);
   } else {
      timer->next  = NULL;
   }
   timer->state = TIMER_UNUSED  | TIMER_TYPE_FREE;
} // freeTimer(struct timer_t)


//--------------------------------------------------------------------

/* Format a time as duration
 *
 *  This function formats a time structure \ref datdur_t as duration.<br />
 *  The format will be \code "1234d 23h59m59s"\endcode. 16 characters will
 *  be put to \c *s including a trailing zero.
 *
 *  \param s     start position within string resp. char array
 *  \param tRun  pointer to the time structure
 *  \return points to the character behind (i.e. the trailing zero)
 */
char * formTimDur(char * s, datdur_t * tRun){
//  0123456789.12345
// "....d ..h..m..s";
  fourDigs(s, tRun->d);
  twoDigs(&(s[6]), tRun->h);
  twoDigs(&(s[9]), tRun->m);
  twoDigs(&(s[12]), tRun->s);
  s[5] = ' ';
  s[4] = 'd';
  s[8] = 'h';
  s[11] = 'm';
  s[14] = 's';
  s += 15;
  *s = 0;
  return s;
}  // formTimDur

/*  Format a time (datdur_t structure) as time of day
 *
 *  This function formats a time structure \ref datdur_t as time of day or
 *  clock time. <br />
 *  The format will be \code "23:59:59"\endcode. 9 characters will
 *  be put to \c *s including a trailing zero.
 *
 *  The field \c d of \c tTim is not used for the result.
 *
 *  \param s     start position within string resp. char array
 *  \param tTim  pointer to the time structure
 *  \return points to the (0-) character behind the formated duration
 */
char * formTimOfD(char * s, datdur_t * tTim){
   //  0123456789
   // "hh:mm:ss";
     twoDigs(&(s[0]), tTim->h);
     twoDigs(&(s[3]), tTim->m);
     twoDigs(&(s[6]), tTim->s);
     s[2] = s[5] = ':';
     s += 8;
     *s = 0;
     return s;
} // formTimOfD(char *, datdur_t *)


/*  Parse a (clock) time
 *
 *  This function tries to interpret a part of the string s starting at
 *  si as a (clock) time. The expected format is "hh:mm:ss" or "hh:mm" with
 *  the seconds implied as 0. Returned is the index of the first character
 *  after that (parsed) time in case of success and 0 otherwise.
 *
 *  Leading blanks will be skipped. Leading zeroes don't of course change
 *  the result and they do not lead to octal interpretation.
 *
 *  \param tim pointer to the result structure; may be changed partially
 *             even if the parse was not successful
 *  \param s   the RAM string containing the time
 *  \param si  the first time's index in s
 *  \return    the index of the first character after the parsed time
 *             in case of success or 0 else
 */
uint8_t parseTim(datdur_t * tim, char s[], uint8_t si){
   if (!tim) return 0;
   uint8_t resP;
   si = parseByteNum(&resP, s, si);
   if (!si  || resP > 23) return 0; // no hour
   uint8_t c = s[si];
   while (c == ' ') c = s[++si]; // skip trailing blanks
   if (c != ':') return 0; // nothing follows
   tim->h = resP; // hour correct

   si = parseByteNum(&resP, s, si + 1);
   if (!si || resP > 59) return 0; // no minute
   tim->m = resP; // minute correct
   tim->s = 0;
   c = s[si];
   while (c == ' ') c = s[++si]; // skip trailing blanks
   if (c != ':') return si; // have hour:min    OK (imply h:m:0)
   // have h:m:   must have second now
   si = parseByteNum(&resP, s, si + 1);
   if (!si || resP > 60) return 0; // no second
   tim->s = resP; // second correct
   return si;
} //  parseTim(datdur_t *...)


/*  Structure a seconds based (32 bit) time
 *
 *  This function converts a 32 bit unsigned seconds to a time structure
 *  \ref datdur_t. Its interpretation as duration or absolute date and time
 *  depends on the parameter's semantic.
 *
 *  \param  timStr  pointer to the time structure to be set
 *  \param  timSec  the time in seconds
 */
void sec2datdur(datdur_t * timStr, uint32_t timSec){
   if (!timStr) return;

   uint32_t leRem;
   timStr->d = (uint16_t)div32by24(&leRem, timSec,  SECONDS_IN_DAY);

   //udiv32(&accuD32, timSec,  SECONDS_IN_DAY);
   //timStr->d = accuD32.quot.v16[0];
   //udiv32(&accuD32, accuD32.rem.v32,  3600);

   timStr->h  = (uint8_t) div32by24(&leRem, leRem,  3600);
   // timStr->h = accuD32.quot.v8[0];

   // udiv16(&accuD16, accuD32.rem.v16[0], 60);
   timStr->m = div16((uint16_t*)(&leRem), (uint16_t)leRem, 60);
   timStr->s = (uint8_t)leRem;
   //timStr->s = accuD16.rem;
} // sec2datdur(datdur_t, uint32_t)


/*  Calculate seconds from a datdur_t structure
 *
 *  This function converts the content of a structure of type \ref datdur_t
 *  to the corresponding 32 bit unsigned seconds value. It is the inverse of
 *  \ref sec2datdur().
 *
 *  \param  timStr  pointer to the time structure to be used
 *  \return the seconds calculated from timStr
 */
uint32_t datdur2sec(datdur_t * timStr){
   if (!timStr) return 0;
   return mul16with17(timStr->d, (uint16_t)SECONDS_IN_DAY)
                 +  mul16with8(timStr->h * 60 + timStr->m, 60) + timStr->s;
} // datdur2sec(datdur_t)

/* Format a date (date_t structure)
 *
 *  This function formats a date (structure \ref date_t).<br />
 *  The format (ISO 8601 / EN 28601) will be \code "2009-01-31"\endcode.<br />
 *  11 characters will be put \c *s including a trailing zero.
 *
 *  The field \c wd is not used for the result.
 *
 *  \param s     start position within string resp. char array
 *  \param tDat  pointer to the time structure
 *  \return points to the (0-) character behind the formated day of week
 */
char * formDateIso(char * s, date_t * tDat){
   //  0123456789.
   // "2010-01-31";
   int8_t y = tDat->y;
   if (y < 100) {
      s[1] = '0';
   } else if (y < 200) {
      s[1] = '1';
      y -= 100;
   } else {
      s[1] = '2';
      y -= 200;
   }
     twoDigs(&(s[2]), y);
     twoDigs(&(s[5]), tDat->m);
     twoDigs(&(s[8]), tDat->d);
     s[0] = '2';
     s[4] = s[7] = '-';
     s += 10;
     *s = 0;
     return s;
} // formDateIso(char *, dat_t *)

INFLASH(char DoWsh[]) = "DoW ";
INFLASH(char SunSh[]) = "Sun ";
INFLASH(char MonSh[]) = "Mon ";
INFLASH(char TueSh[]) = "Tue ";
INFLASH(char WedSh[]) = "Wed ";
INFLASH(char ThuSh[]) = "Thu ";
INFLASH(char FriSh[]) = "Fri ";
INFLASH(char SatSh[]) = "Sat ";


INFLASH(char* wDaysShort[]) = { DoWsh,    // 0 no day of week just marker
   SunSh, MonSh, TueSh, WedSh, ThuSh, FriSh, SatSh, NULL };

/*  Format the day of week as short clear text
 *
 *  This function puts the three character short English day of week into
 *  \c s. <br />
 *  The format for \c wd 1..7 will be
 *  \code "Sun" "Mon" "Tue" "Wed" "Thu" "Fri" "Sat" \endcode. <br />
 *  4 characters will be put \c *s including a trailing zero.
 *
 *  For \c wd outside 0..7 sust \s *s will be set 0 and \c s is returned.
 *
 *  \param s     start position within string resp. char array
 *  \param wd    1..7 is Sunday .. Saturday
 *  \return points to the (0-) character behind the formated day of week
 */
char * formWdShort(char * s, uint8_t wd){
   if (wd && wd <= 7) {
      memcpy_P(s, (char*)pgm_read_word(&wDaysShort[wd]), 3);
      s += 3;
   }
   *s = 0;
   return s;
} // formWdShort(char *, uint8_t)

INFLASH(uint16_t monOffs[]) = {
      APRIL_OFF,     MAY_OFF,       JUNE_OFF,     JULY_OFF,
      AUGUST_OFF,    SEPTEMBER_OFF, OCTOBER_OFF,  NOVEMBER_OFF, DECEMBER_OFF,
      JANUARY_OFF,   FEBRUARY_OFF };

/*  Get year starting March (includes next January and February)
 *
 *  \param  ds      days since \ref intro_sec "weAutSys"'s day 0
 *  \param  daysInYear pointer to extra result: the the days
 *          in the respective year (0 = March 1st); if NULL this information
 *          is lost
 *  \return the resulting year 8 .. = 2008 ..
 *  \sa datdur2sec
 */
uint8_t getMarchYearByDays(uint16_t * daysInYear, uint16_t ds){
   uint8_t retYear;
   if (ds >= MARCH_2100) { // >= 1.3.2100 not 29.02.2100
      ds -=  MARCH_2100; // make 1.3.2100 (pseudo) zero
      retYear = 100;
   } else { // below 1.3.2100
      retYear = 8;
   } // 1.3.2008 is zero

   uint16_t leRem; // rem:d
   uint8_t const l4y = (uint8_t)div16(&leRem, ds, DAYS_IN4YEARS); // qt: 4y
   retYear += l4y * 4; // uncorrected year

   if (leRem>= (DAYS_IN_YEAR * 2)) {
      leRem -= (DAYS_IN_YEAR * 2);
      retYear += 2;
   }
   if (leRem >= (DAYS_IN_YEAR)) {
      leRem -= (DAYS_IN_YEAR);
      retYear += 1;
   }
   // correct year (yet starting march)
   if (daysInYear) *daysInYear = leRem;
   return retYear;
} // getMarchYearByDays()


/*  Set date structure by days since since March 2008
 *
 *  This function sets the (16 bit unsigned) number of days since March 2008
 *  into the structure datStr and all its other fields.
 *
 *  \param  datStr  pointer to the date structure to be set
 *  \param  ds      days since \ref intro_sec "weAutSys"'s day 0
 *  \return the resulting weekday (1..7) or 0 in erroneous situations
 */
uint8_t setDatByDays(date_t * datStr, uint16_t ds){
   if (!datStr) return 0;
   datStr->wd =  mod16byVal7(ds);  //
   if (! datStr->wd) datStr->wd = 7; // day of week

   // u16div_t divTemp;
   uint16_t leRem; // rem:d

   datStr->y = getMarchYearByDays(&leRem, ds);

   if (leRem >= JANUARY_OFF ) { // January or February next year
      datStr->y += 1;
      datStr->d = (uint8_t)(leRem - JANUARY_OFF);
      if (datStr->d >= 31) {
         datStr->d -= 31;
         datStr->m = 2;
      } else {
         datStr->m = 1;
      }
   }  else {   // March .. December this year
      datStr->m = 3; // march
      uint16_t * monOffP = monOffs;
      uint16_t actMonOff = 0;
      uint16_t monOff = pgm_read_word(monOffP);

      while (leRem >= monOff) {  // linear search in sorted list
         actMonOff = monOff;
         if (++datStr->m == 12) break;
         monOff = pgm_read_word(++monOffP);
      }  // todo linear search in sorted list may be improved
      datStr->d = (uint8_t)(leRem - actMonOff);
   }
   ++ datStr->d; // day in month  N.b.: all above days counting was 0 based

   return datStr->wd;
} // setDatByDays(date_t, uint16_t)


INFLASH(uint16_t monOffsMod7[]) = {
      APRIL_OFF % 7,    MAY_OFF % 7,       JUNE_OFF % 7,   JULY_OFF % 7,
      AUGUST_OFF % 7,   SEPTEMBER_OFF % 7, OCTOBER_OFF % 7,
      NOVEMBER_OFF % 7, DECEMBER_OFF % 7,
      JANUARY_OFF % 7,  FEBRUARY_OFF  % 7};

/*  Parse a date
 *
 *  This function tries to interpret a part of the string \c s starting at
 *  \c si as a (clock) time. The expected format is "yyyy-mm-dd" or
 *  "yy-mm-dd" with two digit year interpreted as 2008 .. 2107 which is more
 *  than the recommended / allowed range of years.
 *
 *  Leading and trailing blanks at every segment will be skipped as will be
 *  (allowed) leading zeroes. Hence "12-18-1" and " 2012 - 18 - 01 " will get
 *  the same interpretation.
 *
 *  \param dat pointer to the result structure; may be changed partially
 *             even if the parse was not successful
 *  \param s   the RAM string containing the date
 *  \param si  the first index of the date in s
 *  \return    the index of the first character after the parsed time
 *             in case of success or 0 else
 */
uint8_t parseDate(date_t * dat, char s[], uint8_t si){
   if (!dat) return 0;
   dat->wd = 0; // mark as unknown yet

   uint16_t resP;
   si = parseWordNum(&resP, s, si);
   if (!si) return 0; // no year
   if (resP > 2000) {
      resP -= 2000;
      if (resP > 169) return 0; // illegal year;
      dat->y = (uint8_t) resP; // year correct
   } else {
      if (resP > 169) return 0; // illegal year;
      dat->y = (uint8_t) resP; // year correct (except 0..7)
      if (!(dat->y & 0xF8)) dat->y += 100; //  0..7 -> + (2)100
   }
   uint8_t c = s[si];
   while (c == ' ') c = s[++si]; // skip trailing blanks
   if (c != '-') return 0; // nothing follows

   si = parseByteNum(&c, s, si + 1);
   if (!si || ! c || c > 12) return 0; // no month
   uint8_t m = dat->m = c; // month correct (1..12)

   c = s[si];
   while (c == ' ') c = s[++si]; // skip trailing blanks
   if (c != '-') return 0; // nothing follows
   si = parseByteNum(&c, s, si + 1);
   if (!si || ! c || c > 31) return 0; // illegal day
   dat->d = c; // day correct

   // below calc. wd
   uint8_t y = dat->y - 8;
   if (m < 3) {
      m += 9;
      --y;
   } else {
      m -= 3;
   }
   if (y >= 92) {
      dat->wd = y - 86 +  y / 4; // 5 - 91 == 5 - 13 * 7
   } else {
      dat->wd = 6 + y + y / 4;  // max 120
   }
   if (m){
      dat->wd += pgm_read_word(&monOffsMod7[m-1]);
   }
   dat->wd = mod8byVal7(dat->wd + dat->d);
   if (!dat->wd) dat->wd = 7;

   return si;
} // parseDate(date_t * ..)

/*  Get days since since March 2008 by date structure
 *
 *  This calculates the (16 bit unsigned) number of days since March 2008
 *  from the structure and returns the result. All fields of datStr except
 *  day of week \c (ed) are used and must have correct and consistent values.
 *
 *  Before March 2008  0 is returned as is for everything after about ~2177.
 *
 *  \param  datStr  pointer to the date structure to be used
 *  \return     days since \ref intro_sec "weAutSys"'s day 0
 */
uint16_t getDaysByDat(date_t * datStr){
   if (!datStr) return 0;
   uint8_t y = datStr->y - 8;
   uint8_t m = datStr->m;
   if (m < 3) {
      m += 9;
      --y;
   } else {
      m -= 3;
   }
   if (y > 170) return 0;
   uint16_t res = datStr->d -1;
   if (y & 3) {
      res += (y & 3) * DAYS_IN_YEAR;
   }
   if (y >>= 2) {
      res += y * DAYS_IN4YEARS;
   }
   if (y >= 92) --res; // 2100 no leap year
   if (m){
      res += pgm_read_word(&monOffs[m-1]);
   }
   return res;
} // getDaysByDat(date_t)

/*  Is date represented in date structure EU DST switching day
 *
 *  This function checks if the date is either the last Sunday in March
 *  (returns 3) or the last Sunday in October (returns 10) or non of
 *  both (returns 0).
 *
 *  \param  datStr  pointer to the date structure
 *  \return the resulting 3, 10 or 0
 */
uint8_t isEUdstSwitchDay(date_t * datStr) {
   if (!datStr) return 0;
   if (datStr->wd != 1) return 0;
   if (datStr->m != 3 && datStr->m != 10) return 0;
   if (datStr->d < 24) return 0;
   return datStr->m;
} // isEUdstSwitchDay(date_t)

/* Lookup table for DST rule number by year 0..255 as 2000 ..2255 */
INFLASH(uint8_t ruleNoByY[]) = {
      1, 0, 6, 5, 3, 2, 1, 0, 5, 4, 3, 2, 0, 6, 5, 4, 2, 1, 0, 6,
      4, 3, 2, 1, 6, 5, 4, 3, 1, 0, 6, 5, 3, 2, 1, 0, 5, 4, 3, 2,
      0, 6, 5, 4, 2, 1, 0, 6, 4, 3, 2, 1, 6, 5, 4, 3, 1, 0, 6, 5,
      3, 2, 1, 0, 5, 4, 3, 2, 0, 6, 5, 4, 2, 1, 0, 6, 4, 3, 2, 1,
      6, 5, 4, 3, 1, 0, 6, 5, 3, 2, 1, 0, 5, 4, 3, 2, 0, 6, 5, 4,
      3, 2, 1, 0, 5, 4, 3, 2, 0, 6, 5, 4, 2, 1, 0, 6, 4, 3, 2, 1,
      6, 5, 4, 3, 1, 0, 6, 5, 3, 2, 1, 0, 5, 4, 3, 2, 0, 6, 5, 4,
      2, 1, 0, 6, 4, 3, 2, 1, 6, 5, 4, 3, 1, 0, 6, 5, 3, 2, 1, 0,
      5, 4, 3, 2, 0, 6, 5, 4, 2, 1, 0, 6, 4, 3, 2, 1, 6, 5, 4, 3,
      1, 0, 6, 5, 3, 2, 1, 0, 5, 4, 3, 2, 0, 6, 5, 4, 2, 1, 0, 6,
      5, 4, 3, 2, 0, 6, 5, 4, 2, 1, 0, 6, 4, 3, 2, 1, 6, 5, 4, 3,
      1, 0, 6, 5, 3, 2, 1, 0, 5, 4, 3, 2, 0, 6, 5, 4, 2, 1, 0, 6,
      4, 3, 2, 1, 6, 5, 4, 3, 1, 0, 6, 5, 3, 2, 1, 0   };

/* DST rule number 0 */
INFLASH(dst_rule_year_t dstRule0) = {
      25, 28, 2077200,  20826000, 11, 4,  871200,   21434400 };
INFLASH(dst_rule_year_t dstRule1) = {
      26, 29, 2163600,  20912400, 12, 5,  957600,   21520800 };
INFLASH(dst_rule_year_t dstRule2) = {
      27, 30, 2250000,  20998800, 13, 6,  1044000,  21607200 };
INFLASH(dst_rule_year_t dstRule3) = {
      28, 31, 2336400,  21085200, 14, 7,  1130400,  21693600 };
INFLASH(dst_rule_year_t dstRule4) = {
      29, 25, 2422800,  20566800, 8,  1,  612000,   21175200  };
INFLASH(dst_rule_year_t dstRule5) = {
      30, 26, 2509200,  20653200, 9,  2,  698400,   21261600 };
INFLASH(dst_rule_year_t dstRule6) = {
      31, 27, 2595600,  20739600, 10, 3,  784800,   21348000 };

/* Lookup table for DST rule address (in Flash) by rule number */
INFLASH(dst_rule_year_t const *  dstRuleByNo[]) = {
      &dstRule0,  &dstRule1,  &dstRule2,  &dstRule3,
      &dstRule4,  &dstRule5,  &dstRule6 };


/* Get the EU, US &amp;c. DST rule for a given year
 *
 *  This function returns a pointer to the
 *  \ref  dst_rule_year_t "DST rule structure" for the year requested in
 *  flash memory. To have it in RAM use \code
 *
 * \param year 0..255 interpreted as 2000..2255
 */
const dst_rule_year_t  *  getDSTrule(uint8_t year) {
   year = pgm_read_byte(&ruleNoByY[year]);  // rule no by year in flash
   return (dst_rule_year_t  *)(pgm_read_word(&dstRuleByNo[year]));
} // getDSTrule(uint8_t year)



// internal: moved here from ntp.c due to the damned
// inline asm bug 48435
#include "we-aut_sys/ntp.h"

/* Milliseconds from a NTP time stamp fraction
 *
 *  This function converts a NTP time stamp fraction value into (correct byte
 *  order) milliseconds in the range 0..999
 *
 *  \param value points to the time stamp to be converted (not NULL)
 *  \return value.fraction converted to ms (0 .. 999)
 */
void ntpTimestampToMillies(uint16_t * ms, struct ntpTimestamp_t * ntpStamp){

    asm volatile(""
          "LDD  r24, Y+4   \n\t"
          "MOV  __tmp_reg__, r24  \n\t"
          "LDD  r25, Y+5   \n\t"
          "andi r25, 0xC0 \n\t"
          "eor  __zero_reg__ , __zero_reg__ ; hope it helps \n\t"
          "lsl  r24   \n\t"
          "rol  r25   \n\t"
          "adc  r24, __zero_reg__  \n\t"
          "lsl  r24   ; two left (+ byte resort) is  ten to left  \n\t"
          "rol  r25   ; i.e.  * 1024 en lieu  \n\t"
          "adc  r24, __zero_reg__  \n\t"

          "lsr  __tmp_reg__ ; by >> 3 is in effect  \n\t"
          "lsr  __tmp_reg__ ; (with << 2 above) by 5 \n\t"
          "lsr  __tmp_reg__ ; 1/32 ~= 1/40 i.e 1-1024/100 correction  \n\t"
          "sub r24, __tmp_reg__   \n\t"
          "sbc r25, __zero_reg__  \n\t"
          "st  z, r24    \n\t"
          "std z+1, r25    \n\t"
          :
          : "z" (ms), "y" (ntpStamp)
          : "r24", "r25"
    );

} // ntpTimestampToMillies(uint16_t *, struct ntpTimestamp_t *)


