// Copyright (c) 2011 Albrecht Weinert, Bochum *
// $Id: uart0.c 2 2017-01-25 16:49:03Z albrecht $

/**
 * \addtogroup weAutSys
 * @{
 * \addtogroup sysSer
 * @{
 * \addtogroup serCom
 * @{
 *
 * \file uart0.c
 *
 *  Serial communication (basics)
 *
 *  This file contains the implementations for the basics of weAutSys' serial
 *  communication.
 *  This is system software and must not be modified for user or
 *  application programs.
 *
 *
 * Please find detailed description of the variables, functions etc. in the
 * module \ref serCom "Serial communication" and in the include file uart0.h.
 *
 * This file is part of weAutSys &nbsp;
 *           <<a href="http://weinert-automation.de">weinert-automation.de</a>>
 *
 * &nbsp; \ref copyLicNote " Copyright &copy; " 2011 Albrecht Weinert, Bochum
 * \author  Albrecht Weinert  &nbsp; <<a href="http://a-weinert.de">a-weinert.de</a>>
 * $Revision: 2 $
 * $Date: 2017-01-25 16:49:03 +0000 (Wed, 25 Jan 2017) $
 */
//  SVN last change by $Author: albrecht $


/** @} */ // doxygen addtogroup serCom
/** @} */ // doxygen addtogroup sysSer
/** @} */ // doxygen addtogroup weAutSys

// the next two includes are rubbish for Eclipse Indigo .h troubles only
//#include <avr/io.h>
//#undef _AVR_IOXXX_H_
//#  include <avr/iom1284p.h>
// delete three # above as soon as Indigo can handle nested .h files

#include <util/atomic.h>

#include "we-aut_sys/common.h"
#include "we-aut_sys/uart0.h"

#include "we-aut_sys/system.h"
#include "we-aut_sys/syst_threads.h"
#include "we-aut_sys/ll_system.h"
#include "we-aut_sys/utils.h"
#include "we-aut_sys/timing.h"

/* The serial output buffer */
uint8_t uartOutBuf[UART_OUT_BUF_CAP + 1];

/* The serial output buffer write/put index
 *
 * It is volatile because multiply written in in one protothread slot,
 * while the  interrupt handler may read.
 */
volatile uint8_t uartOutBufWi;

/* The serial output buffer read/get index
 *
 * Must be volatile, as it may be accessed more than once in one protothread
 * slot, while interrupt handler may modify.
 */
volatile uint8_t uartOutBufRi;

/* Clear the internal buffer for serial output */
uint8_t uartClearOutBuffer(){
   uint8_t filled;
   ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
      filled = (uartOutBufWi - uartOutBufRi) &  UART_OUT_BUF_CAP;
      uartOutBufRi = uartOutBufWi = 0;
   }  // atomic block
   return filled;
} // uartClearOutBuffer()


/* The buffer space available for serial output */
uint8_t uartOutSpace(){
   uint8_t space;
   ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
      space =  (UART_OUT_BUF_CAP + uartOutBufRi - uartOutBufWi)
             &  UART_OUT_BUF_CAP;
   }  // atomic block
   return space;
} // uartOutSpace()


/* The buffer space available for serial output */
uint16_t serStreamsOutSpace(FILE * streams){
   // inlined uartOutSpace()
   uint8_t space;
   ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
      space =  (UART_OUT_BUF_CAP + uartOutBufRi - uartOutBufWi)
             &  UART_OUT_BUF_CAP;
   }  // atomic block
   return space;
} //  serStreamsOutSpace()


/*  Check the space available for standard output
 *
 *  This function returns \ref checkUARToutSpace() if the stdout
 *  \ref SER_IS_STDOUT "is the" UART, \ref checkLANoutSpace()
 *  if the stdout
 *  \ref LAN_IS_STDOUT "is the" LAN stream and 1 in all other cases.
 *
 *  \param reqSpace the number of bytes requested for output in one complete
 *         step
 *  \return 0: is available. I.e. go ahead with planned output.<br />
 *       1: may be available, but can't be determined for
 *       the current (standard) output device. I.e. go ahead or give up;<br />
 *       2: is not available, but may become (automatically in the future,
 *          i.e. yielding / waiting is OK;<br />
 *       3: is not available, but may become by own activity (flushing,
 *          sending). I.e. just yielding / waiting would be endless.
 *       4: will never be available. I.e. give up or request less space.
 */
uint8_t serCheckOutSpace(uint16_t reqSpace, FILE * streams){
   if (reqSpace > UART_OUT_BUF_CAP) return 4; // to large : never
   uint8_t space;
   ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
      space = (UART_OUT_BUF_CAP + uartOutBufRi - uartOutBufWi)
                       &  UART_OUT_BUF_CAP;
   }  // atomic block
   return reqSpace > space ? 2 : 0;
} // serCheckOutSpace(uint16_t)


/* Put one byte to serial output
 *
 * This function will be used by stdio as output driver entry.
 * (Hence its necessity and the unused FILE*-parameter.)
 */
int serPutChar(char c, FILE *stream){
   ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
      // setTestPin0(ON); // measure char time

      uint8_t space = (UART_OUT_BUF_CAP - uartOutBufWi + uartOutBufRi)
                                                        &  UART_OUT_BUF_CAP;
      if (space == 0) return EOF; // or other error code
      // by ingeniousness of atomic.h this break out is legal

      // any character scrambling for stdio like Unix' / Linux' horrible
      // double line spaces would have do be done here.  But as our
      // customers / users may very well want to transfer binary data
      // over V.24 we refrain -- as all later maintainers should.

      // direct write through to ready UART -- why not
      if (SER_SENDBUF_EMPTY  // the buffer (array) is empty and ...?..
            && UART_CAN_SEND) { // so is the UART transmitter
         UDR0 = c;
         // setTestPin0(OFF); // measure char time
         return c;  // by ingeniousness of atomic.h this break out is legal
      } // all was empty, the fastest case

      // if something buffered and send possible do so w/o interrupt
      if (SER_SENDBUF_NOT_EMPTY  // the buffer (array) is not empty and the
            && UART_CAN_SEND) { // UARTs transmitter register is
         FROM_BUFFER2UART()
      } // something to send anyway

      uartOutBuf[uartOutBufWi] = c;
      uartOutBufWi = (uartOutBufWi + 1 ) & UART_OUT_BUF_CAP;
   }  // atomic block

   //  setTestPin0(OFF); // measure char time
   return c;
} // serPutChar(char, FILE *)



/* Put some characters to serial output  */
uint8_t uartPutChars(char * src, uint8_t n){
   if (!n) return 0;
   char c = *src;
   if (!c) return 0;
   uint8_t retVal = 0;
   uint8_t space  = 0;
   ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
      space = (UART_OUT_BUF_CAP - uartOutBufWi + uartOutBufRi)
                                                        &  UART_OUT_BUF_CAP;
      if (space == 0) return 0;
      if (UART_CAN_SEND) { // the UART transmitter is empty
         if (SER_SENDBUF_EMPTY) {  // the buffer (array) is empty
            UDR0 = c; // send first direct
            if (n == 1) return 1; // and only (special case)
            ++ retVal;
            c = *(++src);
         } else { // buffer not empty
            FROM_BUFFER2UART() // take one from buffer
            ++ space;
         }
      } // UART can send
   } // first atomic block did send at least one byte

   // now just copy remaining to buffer (interrupt may be on, usually is)
   while (retVal < n && c ) {
      uartOutBuf[uartOutBufWi] = c;
      uartOutBufWi = (uartOutBufWi + 1 ) & UART_OUT_BUF_CAP;
      if (!(--space)) break; // while
      ++retVal;
      c = *(++src);
   } // while

   ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
      if (UART_CAN_SEND && SER_SENDBUF_NOT_EMPTY ) {
         FROM_BUFFER2UART() // get or keep it going
      }
   }
   return retVal;
} // uartPutChars(char *, uint8_t)

/*  Put some bytes (characters) to serial output
 *
 *  \param src the characters to be output
 *  \param n   the maximum number of characters to be output
 *  \return the number of characters output
 */
uint8_t uartPutBytes(uint8_t * src, uint8_t n){
   if (!n || !src) return 0;
   char c = *src;
   uint8_t retVal = 0;
   uint8_t space  = 0;
   ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
      space = (UART_OUT_BUF_CAP - uartOutBufWi + uartOutBufRi)
                                                        &  UART_OUT_BUF_CAP;
      if (space == 0) return 0;
      if (UART_CAN_SEND) { // the UART transmitter is empty
         if (SER_SENDBUF_EMPTY) {  // the buffer (array) is empty
            UDR0 = c; // send first direct
            if (n == 1) return 1; // and only (special case)
            ++ retVal;
            c = *(++src);
         } else { // buffer not empty
            FROM_BUFFER2UART() // take one from buffer
            ++ space;
         }
      } // UART can send
   } // first atomic block did send at least one byte

   // now just copy remaining to buffer (interrupt may be on, usually is)
   while (retVal < n) {
      uartOutBuf[uartOutBufWi] = c;
      uartOutBufWi = (uartOutBufWi + 1 ) & UART_OUT_BUF_CAP;
      if (!(--space)) break; // while
      ++retVal;
      c = *(++src);
   } // while

   ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
      if (UART_CAN_SEND && SER_SENDBUF_NOT_EMPTY ) {
         FROM_BUFFER2UART() // get or keep it going
      }
   }
   return retVal;
} // uartPutBytes(char *, uint8_t)


/*  Put a RAM string (some characters) to serial output
 *
 *  \param src the characters to be output (not NULL !)
 *  \return the number of characters output
 */
uint8_t uartPutSt(char * src, const FILE * stream){
   if (!src) return 0;
   char c = *src;
   if (!c) return 0;
   uint8_t retVal = 0;
   uint8_t space  = 0;
   ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
      space = (UART_OUT_BUF_CAP - uartOutBufWi + uartOutBufRi)
                                                        &  UART_OUT_BUF_CAP;
      if (space == 0) return 0;
      if (UART_CAN_SEND) { // the UART transmitter is empty
         if (SER_SENDBUF_EMPTY) {  // the buffer (array) is empty
            UDR0 = c; // send first direct
           ///  if (n == 1) return 1; // and only (special case)
            ++ retVal;
            c = *(++src);
         } else { // buffer not empty
            FROM_BUFFER2UART() // take one from buffer
            ++ space;
         }
      } // UART can send
   } // first atomic block did send at least one byte

   // now just copy remaining to buffer (interrupt may be on, usually is)
   while (c) {
      uartOutBuf[uartOutBufWi] = c;
      uartOutBufWi = (uartOutBufWi + 1 ) & UART_OUT_BUF_CAP;
      if (!(--space)) break; // while
      ++retVal;
      c = *(++src);
   } // while

   ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
      if (UART_CAN_SEND && SER_SENDBUF_NOT_EMPTY ) {
         FROM_BUFFER2UART() // get or keep it going
      }
   } // ATOMIC_BLOCK
   return retVal;
} //  uartPutSt(char *, const FILE *)


/*  Put a string (some characters) from flash memory to serial output
 *
 *  \param src the characters to be output (not NULL !)
 *  \return the number of characters output
 */
int uartPutSt_P(prog_char * src, const FILE * stream){
   if (!src) return 0;
   char c = pgm_read_byte(src);
   if (!c) return 0;
   uint8_t retVal = 0;
   uint8_t space  = 0;
   ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
      space = (UART_OUT_BUF_CAP - uartOutBufWi + uartOutBufRi)
                                                          &  UART_OUT_BUF_CAP;
      if (space == 0) return 0;
      if (UART_CAN_SEND) { // the UART transmitter is empty
         if (SER_SENDBUF_EMPTY) {  // the buffer (array) is empty
            UDR0 = c; // send first direct
            //  if (n == 1) return 1; // and only (special case)
            ++ retVal;
            c = pgm_read_byte(++src);
         } else { // buffer not empty
            FROM_BUFFER2UART() // take one from buffer
            ++ space;
         }
      } // UART can send
   } // first atomic block did send at least one byte

   // now just copy remaining to buffer (interrupt may be on, usually is)
   while (c) {
      uartOutBuf[uartOutBufWi] = c;
      uartOutBufWi = (uartOutBufWi + 1 ) & UART_OUT_BUF_CAP;
      if (!(--space)) break; // while
      ++retVal;
      c = pgm_read_byte(++src);
   } // while

   ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
      if (UART_CAN_SEND && SER_SENDBUF_NOT_EMPTY ) {
         FROM_BUFFER2UART(); // get or keep it going
      }
   }

   return retVal;
} // uartPutSt_P(prog_char *, const FILE * )


/* Put some characters from program space to serial output */
int uartPutCharsP(char const * src, uint8_t n){
   if (!n) return 0;
   char c = pgm_read_byte(src);
   if (!c) return 0;
   uint8_t retVal = 0;
   uint8_t space  = 0;
   ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
      space = (UART_OUT_BUF_CAP - uartOutBufWi + uartOutBufRi)
                                                        &  UART_OUT_BUF_CAP;
      if (space == 0) return 0;
      if (UART_CAN_SEND) { // the UART transmitter is empty
         if (SER_SENDBUF_EMPTY) {  // the buffer (array) is empty
            UDR0 = c; // send first direct
            if (n == 1) return 1; // and only (special case)
            ++ retVal;
            c = pgm_read_byte(++src);
         } else { // buffer not empty
            FROM_BUFFER2UART() // take one from buffer
            ++ space;
         }
      } // UART can send
   } // first atomic block did send at least one byte

   // now just copy remaining to buffer (interrupt may be on, usually is)
   while (retVal < n && c ) {
      uartOutBuf[uartOutBufWi] = c;
      uartOutBufWi = (uartOutBufWi + 1 ) & UART_OUT_BUF_CAP;
      if (!(--space)) break; // while
      ++retVal;
      c = pgm_read_byte(++src);
   } // while

   ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
      if (UART_CAN_SEND && SER_SENDBUF_NOT_EMPTY ) {
         FROM_BUFFER2UART() // get or keep it going
      }
   } // ATOMIC_BLOCK

   return retVal;
} // uartPutChars(prog_char *, uint8_t)

//-------------------------------------------------------------------

// #define USART_TX_vect     _VECTOR(20)  on 328P

#ifdef USART_RX_vect // we're on a small device with only one UART
#define USART0_RX_vect USART_RX_vect
#define USART0_TX_vect USART_TX_vect
#endif

/** The UART (0) transmission complete interrupt
 *
 *  \internal For (buffered) serial transmission at 38200, 8bit ... default
 *  this will occur every 0.52ms as two bytes are transmitted on one
 *  interrupt whenever possible.
 */
ISR(USART0_TX_vect){
#if USE_TESTPINS == 2
   setTestPin0(ON); // TEST  measure send interrupt
#endif

   if (SER_SENDBUF_NOT_EMPTY  // the buffer (array) is not empty and the
            && UART_CAN_SEND) { // UARTs transmitter register is
      FROM_BUFFER2UART()

      // if one byte can be sent a second one may also be sent
      if (SER_SENDBUF_NOT_EMPTY  // the buffer (array) is not empty and the
                  && UART_CAN_SEND) { // UARTs transmitter register is
           FROM_BUFFER2UART()
      }
   }
#if USE_TESTPINS == 2
   setTestPin0(OFF); // TEST  measure send interrupt
#endif
} // transmission complete interrupt

/* USART0, TX Complete */
//#define USART0_TX_vect        _VECTOR(22)
//#define SIG_USART_TRANS       _VECTOR(22)

// --  input  -----------------------------

static uint8_t b1d6Mode; // usage of the ports PD6 and PB1


/* USART0, RX Complete */
//#define USART0_RX_vect        _VECTOR(20)
//#define SIG_USART_RECV        _VECTOR(20)
// #define USART_RX_vect     _VECTOR(18)  on 328P
// #define USART_TX_vect     _VECTOR(20)  on 328P


/** The UART (0) character received interrupt
 *
 * \internal Must get the character, else re-occurrence of interrupt.
 */
ISR(USART0_RX_vect){
#if USE_TESTPINS == 2
   setTestPin1(ON); // TEST  measure receive interrupt
#endif
   uint8_t status;

   while ( (status = UCSR0A) & (1<<RXC0) ) { // if / while reception
      uint8_t recByte = UDR0;
      uartInAccErrors |= status & 0x1C; // FE DO PE
#if UART0SPIYKY
      if ((recByte & 0xFE) == 0xFE) {
         uartInAccErrors |= 0x20;
         status |= 0x10; // emulate as framing error to filter it
      }
#endif

      if (uartInBufWi == uartInBufRi) { // empty buffer
         appSerInpThread.flag = uartInBufRetCnt =
                                  recByte == LF || recByte == CR ? 1 : 0;
         uartInBufWi = (status & 0x14) ? 0 : 1; // omit FE, PE and spikes
         uartInBufRi = 0;
         uartInBuf[0]= recByte;
         continue; // try to get next if already there
      } // empty buffer

      uint8_t space = (UART_IN_BUF_CAP - uartInBufWi + uartInBufRi)
                                                         & UART_IN_BUF_CAP;
      if (space == 0) {
         uartInAccErrors |= 0x40; // Bit 6
         return;
      }
      if (space == UART_IN_SPACE_LIM) {
         appSerInpThread.flag = 1;
#if UART0MAY_CTS
         if(isCTSflowC) {
            orDP(1<<6); // PD6 (CTS) set to 1 (-12V): be silent
         }
#endif
      }
      if (recByte == LF || recByte == CR) {
         appSerInpThread.flag = 1;
         ++uartInBufRetCnt;
      }
      uartInBuf[uartInBufWi]= recByte;
      if (!(status & 0x14)) { // ignore if FE, PE and spikes
         uartInBufWi = (uartInBufWi + 1) & UART_IN_BUF_CAP;
      } // ignore if FE
   } // if while reception
#if USE_TESTPINS == 2
   setTestPin1(OFF); // TEST  measure receive interrupt
#endif
} //  character received interrupt

/* The serial input buffer */
uint8_t uartInBuf[UART_IN_BUF_CAP + 1];

/* The serial input buffer write/put index  */
uint8_t uartInBufWi;

/* The serial input buffer read/get index  */
uint8_t uartInBufRi;

/* The serial input buffer line feeds counter  */
uint8_t uartInBufRetCnt;

/* The serial input's accumulated (or) errors  */
uint8_t uartInAccErrors;


/* The number of characters buffered from serial input  */
uint8_t uartInBufferd(){
   uint8_t filled;
   ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
      filled = (uartInBufWi - uartInBufRi) & UART_IN_BUF_CAP;
   }  // atomic block
   return filled;
} // uartInBufferd()


/* The number of characters buffered from serial input  */
uint16_t serInBufferd(FILE * streams){
   uint8_t filled;
   ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
      filled = (uartInBufWi - uartInBufRi) & UART_IN_BUF_CAP;
   }  // atomic block
   return filled;
} // serInBufferd(FILE *)

/* The number of line feeds buffered from serial input  */
uint8_t uartInRetBufferd(){
   return uartInBufRetCnt; // consider byte read as atomic
} // uartInRetBufferd()

/* Get one byte from serial input  */
int uartGetChar(FILE *stream){
   uint8_t fromBu;
   ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
      if (uartInBufWi == uartInBufRi) return EOF; // empty
      fromBu = uartInBuf[uartInBufRi];
      uartInBufRi = (uartInBufRi + 1) & UART_IN_BUF_CAP;
      if (fromBu == LF || fromBu == CR)  --uartInBufRetCnt;
#if UART0MAY_CTS
      if (isCTSflowC) {
         uint8_t filled = (uartInBufWi - uartInBufRi) & UART_IN_BUF_CAP;
         if (filled == 10) {
            andDP(~(1<<6)); // PD6 (CTS) set to 0 (+12V): talk ahead
         }
      } // isCTSflowC
#endif
   }  // atomic block
   return fromBu; // dummy
} // uartGetChar(FILE*)


/* Read a line from UART
 *
 * This function will copy all characters to the array (/c line) supplied
 * ending with a line feed, put as LF as last character.
 *
 * Returned is the number of characters copied to the array /c line.
 */
uint8_t uartGetLine(char * line, uint8_t max){
   if (!max) return 0;
   if (!line) return 0;
   //if (max > UART_IN_BUF_CAP) max = UART_IN_BUF_CAP;
   uint8_t put = 0;
   ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
      while (put < max) {
         if (uartInBufWi == uartInBufRi) break; // empty: out of while
         uint8_t fromBu = uartInBuf[uartInBufRi];
         uartInBufRi = (uartInBufRi + 1) & UART_IN_BUF_CAP;
         if (fromBu == CR || fromBu == LF) { // found the ending line feed
            *line++ = 0;
            ++put;
            --uartInBufRetCnt;
            if (uartInBufWi != uartInBufRi) { // not empty
               uint8_t next = uartInBuf[uartInBufRi]; // peek ahead
               if (next == LF || next == CR) { // next is also feed
                  if (next != fromBu) { // forget second of LF CR or CR LF
                     uartInBufRi = (uartInBufRi + 1) & UART_IN_BUF_CAP;
                     --uartInBufRetCnt;
                  }  // forget second of LF CR or CR LF
               }  // next is also feed
            } // not empty
          break; // ready by ending line feed
         } // found the ending line feed
         // if ((fromBu & 0xE0) == 0xE0) continue; // filter (experimental
         *line++ = fromBu;
         ++put;
      } // while
#if UART0MAY_CTS
      if (isCTSflowC) {
         uint8_t filled = (uartInBufWi - uartInBufRi) & UART_IN_BUF_CAP;
         if (filled <= (UART_IN_BUF_CAP - 3 * UART_IN_SPACE_LIM)) {
            andDP(~(1<<6)); // PD6 (CTS) set to 0 (+12V): talk ahead
         }
      }
#endif
   }  // atomic block
   return put;
} // uartGetLine(char[], int)

/* Read a command line from UART
 *
 * This function will copy all characters to the array (/c line) supplied
 * ending with a line feed, put as LF as last character.
 *
 * Returned is the number of characters copied to the array /c line.
 */
uint8_t uartGetCmdLine(char * line, uint8_t max){
   uint8_t put = 0;

   ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
      if (uartInBufWi == uartInBufRi) return put;
      if (!max) return put;
      if (!line) return put;

      uint8_t * riPo = &uartInBuf[uartInBufRi];

      while (uartInBufWi != uartInBufRi) {
         uint8_t fromBu = *(riPo++); // uartInBuf[uartInBufRi];
         if (riPo == &uartInBuf[UART_IN_BUF_CAP]) riPo = &uartInBuf[0];
         uartInBufRi = (uartInBufRi + 1) & UART_IN_BUF_CAP;

         if (fromBu == CR || fromBu == LF) { // found the ending line feed
            *line++ = 0;
            ++put;
            --uartInBufRetCnt;
            if (uartInBufWi != uartInBufRi) { // not empty
               uint8_t next = *riPo; // peek ahead
               if (next == LF || next == CR) { // next is also feed
                  if (next != fromBu) { // forget second of LF CR or CR LF
                     uartInBufRi = (uartInBufRi + 1) & UART_IN_BUF_CAP;
                     --uartInBufRetCnt;
                  }  // forget second of LF CR or CR LF
               }  // next is also feed
            } // not empty
          break; // ready by ending line feed
         } // found the ending line feed
         if (!put) { // filter (experimental)
            if (fromBu <= ' ') continue; // white space, other control
            if (fromBu >  'z') continue; // non ASCII may be due to spikes
          } // filter (experimental)
         *line++ = fromBu;
         if (++put == max) break;
      } // while

#if UART0MAY_CTS
      if (isCTSflowC) {
         uint8_t filled = (uartInBufWi - uartInBufRi) & UART_IN_BUF_CAP;
         if (filled <= (UART_IN_BUF_CAP - 3 * UART_IN_SPACE_LIM)) {
            andDP(~(1<<6)); // PD6 (CTS) set to 0 (+12V): talk ahead
         }
      }
#endif

   }  // atomic block
   return put;
} // uartGetCmdLine(char[], int)


/* Clear the internal buffer for serial input  */
uint8_t uartClearInBuffer(){
   uint8_t filled;
   ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
      filled = ( uartInBufWi - uartInBufRi) &  UART_IN_BUF_CAP;
      appSerInpThread.flag =
            uartInBufWi = uartInBufRi = uartInBufRetCnt = uartInAccErrors = 0;
#if UART0MAY_CTS
      if (isCTSflowC) {
         andDP(~(1<<6)); // PD6 (CTS) set to 0 (+12V): talk ahead
      }
#endif
   }  // atomic block
   return filled;
} // uartClearInBuffer()

/* The accumulated serial input errors  */
uint8_t uartInErrors(){
   uint8_t errors;
  //  ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
      errors = uartInAccErrors;
      uartInAccErrors = 0;
  // }  // atomic block
   return  errors;
} // uartInErrors()

//-- common / initialise --


/*  Initialise the serial input (UART0)  */
void uartInit(uint16_t baudDivide,  uint8_t x2,
                           uint8_t len, uint8_t parity, uint8_t stopBits){
   if (x2) {
    UCSR0A = (1 << U2X0) | (1 << TXC0);
   } else {
    UCSR0A = (1 << TXC0);
   }

   // UCSR0C = was here ... put at the end
   UBRR0 =  baudDivide;  // set baudrate prescaler
   //  uartSetBaudDivide(baudDivide); // set baudrate prescaler

   uartClearOutBuffer();
   uartClearInBuffer();

   UCSR0B = (1 << RXEN0)  | (1 << TXEN0)   // enable receive and transmit +
          | (1 << RXCIE0) | (1 << TXCIE0); // receive + transmit interrupts

      //                              odd    even   none
   uint8_t r0c = parity ? ( (parity & 1) ? 0x30 : 0x20  ) : 0;
   if (stopBits == 2) r0c |= 0x08;
   switch (len) {
   case 5 : break;
   case 6 : r0c |= 0x02; break;
   case 7 : r0c |= 0x04; break;
   case 9 : // not supported fall through default 8
   default: r0c |= 0x06; break;
   }
   UCSR0C = r0c;
} // uartInit(uint16_t, ... )


/* Set the UART0's baudrate divisor  */
void uartSetBaudDivide(uint16_t baudDivide){
  UBRR0 =  baudDivide;
}


#if USE_BOOTLOADER < 2  //---------------------------------------------------

/* Calculate the pre-scaler setting for a desired baud-rate
 *
 * The calculation is based on the the nominal quartz processor clock, i.e.
 * macro F_CPU. Most Arduinos use crystals of so low grade that one has to
 * measure and define it as (macro) shitCrystal in Hz.
 */
uint16_t uartPrescaler(uint32_t baudRate, uint8_t x2){
   baudRate += baudRate; // *2
   baudRate += baudRate; // *4
   baudRate += baudRate; // *8
   if(!x2)    baudRate += baudRate; // *16   (may get 17 bit for 115200)
#if (defined(arduMega2560) || defined(arduinoUno) ) && defined(shitCrystal)
  // set shitCristal to the measured wrong quartz frequency (e.g. 15941400 )
  // return (uint16_t)( ((uint32_t)(shitCrystal))/ (baudRate * (x2 ? 8 : 16) ) );
   // for 16 MHz 24 bit (div24)  would be enough
   return (uint16_t)div32by24(NULL, (uint32_t)(shitCristal), baudRate);
#else
   // FCPU may be 17 bit if 20000000
  // return (uint16_t)( ((uint32_t)(F_CPU))/ (baudRate * (x2 ? 8 : 16) ) );
  return (uint16_t)div32by24(NULL, (uint32_t)(F_CPU), baudRate);
  //  return (uint16_t)( ((uint32_t)(F_CPU)) / baudRate);
#endif
} // uartPrescaler(uint32_t,  uint8_t)


uint32_t uartBaud(uint16_t uartPresc, uint8_t x2){
   uartPresc += 1; // divide by (uartPresc + 1) *  (x2 ? 8 : 16) )
   uartPresc += uartPresc; // (+1)*2
   uartPresc += uartPresc; // ..*4
   uartPresc += uartPresc; //  *8
   if(!x2)    uartPresc += uartPresc; // *16  (may get 16 bit but not larger)
#if (defined(arduMega2560) || defined(arduinoUno)) && defined(shitCrystal)
  // return ((uint32_t)(shitCrystal))
    //     / ((uint32_t)(  (uartPresc + 1) *  (x2 ? 8 : 16) )) ;
   return (uint16_t)div32by16(NULL, (uint32_t)(shitCrystal), uartPresc);
  // set shitCristal to the measured wrong quartz frequency (e.g. 15941400 )
#else
   // return ((uint32_t)(F_CPU)) / uartPresc;
         // ((uint32_t)(  (uartPresc + 1) *  (x2 ? 8 : 16) )) ;
   return (uint16_t)div32by16(NULL, (uint32_t)(F_CPU), uartPresc);
#endif
} // uartBaud(uint16_t,  uint8_t)

#endif  // --- use above from bootloader ----- (copy out of bootLib.c) -----

/* Get the usage of the ports PD6 and PB1 */
uint8_t getB1D6mode(void){ return b1d6Mode; }

/* Set the usage of the ports PD6 and PB1
 *
 *  \param mode the new mode
 *  \return the previous mode
 */
void setB1D6mode(uint8_t mode){
   if (b1d6Mode  == mode) return;
#if defined(weAut_00) || defined(weAut_01)
   uint8_t setTo = mode & D6MODE_MASK;
   if ((b1d6Mode & D6MODE_MASK) != setTo) { // low nibble (D6) mode change
      if(setTo == D6MODE_CTS_FLOWC) {
         // inpBP(1<<1);  // PB1 as input
         outPins(D, 1<<6); // PD6 as output
      }
   } // lo nibble (D6) mode change
   setTo = mode & B1MODE_MASK;
   if ((b1d6Mode & B1MODE_MASK) != setTo) { // high nibble (B1) mode change
      if(setTo == B1MODE_RTS_FLOWC) {
         inpPins(B, 1<<1);  // PB1 as input
         // outpDP(1<<6); // PD6 as output
      }
   } // high nibble (B1) mode change
#endif // for weAut_0x only
   b1d6Mode = mode;
} // setB1D6mode()

/* Sets UART flow control by (one) character
 *
 *  i c: input flow control by CTS <br />
 *  o r: output flow control by RTS <br />
 *  b  : both CTS and RTS
 *  n, all else: no flow control <br />
 */
void setUARTflowcontrolByChar(char c){
   c |= 0x20; // poor man's lower case
   if (c == 'b') {
      c = B1D6MODE_RTSCTS_FLOWC;
   } else if (c == 'i' || c == 'c') {
      c = D6MODE_CTS_FLOWC ;
   } else if (c == 'o' || c == 'r') {
      c = B1MODE_RTS_FLOWC;
   } else {
      c = 0;
   }
   setB1D6mode(c);
} // setUARTflowcontrolByChar(char)

